from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.db.models import Q
from djblets.webapi.errors import DOES_NOT_EXIST, WebAPIError
from djblets.webapi.fields import (BooleanFieldType,
                                   ChoiceFieldType,
                                   DateTimeFieldType,
                                   DictFieldType,
                                   IntFieldType,
                                   ResourceFieldType,
                                   StringFieldType)

from reviewboard.reviews.models import BaseComment
from reviewboard.webapi.base import ImportExtraDataError, WebAPIResource
from reviewboard.webapi.mixins import MarkdownFieldsMixin
from reviewboard.webapi.resources import resources


class BaseCommentResource(MarkdownFieldsMixin, WebAPIResource):
    """Base class for comment resources.

    Provides common fields and functionality for all comment resources.
    """

    added_in = '1.6'

    fields = {
        'id': {
            'type': IntFieldType,
            'description': 'The numeric ID of the comment.',
        },
        'extra_data': {
            'type': DictFieldType,
            'description': 'Extra data as part of the comment. This depends '
                           'on what is being commented on, and may be '
                           'used in conjunction with an extension.',
            'added_in': '2.0',
        },
        'issue_opened': {
            'type': BooleanFieldType,
            'description': 'Whether or not a comment opens an issue.',
        },
        'issue_status': {
            'type': ChoiceFieldType,
            'choices': tuple(BaseComment.ISSUE_STRING_TO_STATUS.keys()),
            'description': 'The status of an issue.',
        },
        'public': {
            'type': BooleanFieldType,
            'description': 'Whether or not the comment is part of a public '
                           'review.',
            'added_in': '2.0',
        },
        'text': {
            'type': StringFieldType,
            'description': 'The comment text.',
            'supports_text_types': True,
            'added_in': '2.0',
        },
        'text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.TEXT_TYPES,
            'description': 'The mode for the comment text field.',
            'added_in': '2.0',
        },
        'timestamp': {
            'type': DateTimeFieldType,
            'description': 'The date and time that the comment was made.',
            'added_in': '2.0',
        },
        'user': {
            'type': ResourceFieldType,
            'resource': 'reviewboard.webapi.resources.user.UserResource',
            'description': 'The user who made the comment.',
            'added_in': '2.0',
        },
    }

    # Common field definitions for create/update requests
    _COMMON_REQUIRED_CREATE_FIELDS = {
        'text': {
            'type': StringFieldType,
            'description': 'The comment text.',
            'supports_text_types': True,
            'added_in': '2.0',
        },
    }

    _COMMON_OPTIONAL_CREATE_FIELDS = {
        'force_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.TEXT_TYPES,
            'description': 'The text type, if any, to force for returned '
                           'text fields. The contents will be converted '
                           'to the requested type in the payload, but '
                           'will not be saved as that type.',
            'added_in': '2.0.9',
        },
        'text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.SAVEABLE_TEXT_TYPES,
            'description': 'The content type for the comment text field. '
                           'The default is ``plain``.',
            'added_in': '2.0',
        },
    }

    _COMMON_OPTIONAL_UPDATE_FIELDS = {
        'force_text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.TEXT_TYPES,
            'description': 'The text type, if any, to force for returned '
                           'text fields. The contents will be converted '
                           'to the requested type in the payload, but '
                           'will not be saved as that type.',
            'added_in': '2.0.9',
        },
        'text': {
            'type': StringFieldType,
            'description': 'The comment text.',
            'supports_text_types': True,
            'added_in': '2.0',
        },
        'text_type': {
            'type': ChoiceFieldType,
            'choices': MarkdownFieldsMixin.SAVEABLE_TEXT_TYPES,
            'description': 'The new content type for the comment text field. '
                           'The default is to leave the type unchanged.',
            'added_in': '2.0',
        },
    }

    # Field definitions for top-level comment create/update requests
    REQUIRED_CREATE_FIELDS = _COMMON_REQUIRED_CREATE_FIELDS

    OPTIONAL_CREATE_FIELDS = dict({
        'issue_opened': {
            'type': BooleanFieldType,
            'description': 'Whether the comment opens an issue.',
            'added_in': '2.0',
        },
    }, **_COMMON_OPTIONAL_CREATE_FIELDS)

    OPTIONAL_UPDATE_FIELDS = dict({
        'issue_opened': {
            'type': BooleanFieldType,
            'description': 'Whether or not the comment opens an issue.',
            'added_in': '2.0',
        },
        'issue_status': {
            'type': ChoiceFieldType,
            'choices': tuple(BaseComment.ISSUE_STRING_TO_STATUS.keys()),
            'description': 'The status of an open issue.',
            'added_in': '2.0',
        },
    }, **_COMMON_OPTIONAL_UPDATE_FIELDS)

    # Field definitions for comment reply create/update requests
    REPLY_REQUIRED_CREATE_FIELDS = dict({
        'reply_to_id': {
            'type': IntFieldType,
            'description': 'The ID of the comment being replied to.',
        },
    }, **_COMMON_REQUIRED_CREATE_FIELDS)

    REPLY_OPTIONAL_CREATE_FIELDS = _COMMON_OPTIONAL_CREATE_FIELDS
    REPLY_OPTIONAL_UPDATE_FIELDS = _COMMON_OPTIONAL_UPDATE_FIELDS

    def serialize_issue_status_field(self, obj, **kwargs):
        return BaseComment.issue_status_to_string(obj.issue_status)

    def has_access_permissions(self, request, obj, *args, **kwargs):
        return obj.is_accessible_by(request.user)

    def has_modify_permissions(self, request, obj, *args, **kwargs):
        return obj.is_mutable_by(request.user)

    def has_delete_permissions(self, request, obj, *args, **kwargs):
        return obj.is_mutable_by(request.user)

    def create_comment(self,
                       review,
                       fields,
                       text,
                       comments_m2m,
                       issue_opened=False,
                       text_type=MarkdownFieldsMixin.TEXT_TYPE_PLAIN,
                       extra_fields={},
                       save=True,
                       **kwargs):
        """Create a comment based on the requested data.

        This will construct a comment of the type represented by the resource,
        setting the issue states, text, extra_data, and any additional fields
        provided by the caller.

        Args:
            review (reviewboard.reviews.models.review.Review):
                The review owning the comment.

            fields (list of unicode):
                The model fields that can be set through the API.

            text (unicode):
                The comment text.

            comments_m2m (django.db.models.ManyToManyField):
                The review's comments relation, where the new comment will
                be added.

            issue_opened (bool, optional):
                Whether this comment opens an issue.

            text_type (unicode, optional):
                The text type for the comment. This defaults to plain text.

            extra_fields (dict, optional):
                Extra fields from the request not otherwise handled by the
                API resource. Any ``extra_data`` modifications from this will
                be applied to the comment.

            save (bool, optional):
                Whether or not to save the field and update ``comments_m2m``.

                If ``False``, the caller is responsible for performing the
                save.

            **kwargs (dict):
                Keyword arguments representing additional fields handled by
                the API resource. Any that are also listed in ``fields`` will
                be set on the model.

        Returns:
            tuple or djblets.webapi.errors.WebAPIError:
            Either a successful payload containing the comment, or an error
            payload.
        """
        comment_kwargs = {
            'issue_opened': bool(issue_opened),
            'rich_text': text_type == self.TEXT_TYPE_MARKDOWN,
            'text': text.strip(),
        }

        for field in fields:
            comment_kwargs[field] = kwargs.get(field)

        new_comment = self.model(**comment_kwargs)

        try:
            self.import_extra_data(new_comment, new_comment.extra_data,
                                   extra_fields)
        except ImportExtraDataError as e:
            return e.error_payload

        if issue_opened:
            new_comment.issue_status = BaseComment.OPEN
        else:
            new_comment.issue_status = None

        if save:
            new_comment.save()
            comments_m2m.add(new_comment)

        return 201, {
            self.item_result_key: new_comment,
        }

    def create_or_update_comment_reply(self, request, comment, reply,
                                       comments_m2m, default_attrs={},
                                       *args, **kwargs):
        """Create a reply to a comment based on the requested data.

        If there's an existing reply to a comment, that one will be updated
        instead.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            comment (reviewboard.reviews.models.base_commet.BaseComment):
                The comment being replied to.

            reply (reviewboard.reviews.models.review.Review):
                The review reply owning the comment.

            comments_m2m (django.db.models.ManyToManyField):
                The reply's comments relation, where the new comment will
                be added.

            default_attrs (dict, optional):
                Default attributes to add to the new comment reply, if an
                existing one does not exist.

            *args (tuple):
                Positional arguments from the caller.

            **kwargs (dict):
                Keyword arguments from the caller.

        Returns:
            tuple or djblets.webapi.errors.WebAPIError:
            Either a successful payload containing the comment, or an error
            payload.
        """
        q = self._get_queryset(request, *args, **kwargs)
        q = q.filter(Q(reply_to=comment) & Q(review=reply))

        try:
            new_comment = q.get()

            # This already exists. Go ahead and update, but we're going to
            # redirect the user to the right place.
            is_new = False
        except self.model.DoesNotExist:
            new_comment = self.model(reply_to=comment, **default_attrs)
            is_new = True

        rsp = self.update_comment(request=request,
                                  review=reply,
                                  comment=new_comment,
                                  is_reply=True,
                                  **kwargs)

        if isinstance(rsp, WebAPIError):
            return rsp

        data = rsp[1]

        if is_new:
            comments_m2m.add(new_comment)
            reply.save()

            return 201, data
        else:
            return 303, data, {
                'Location': self.get_href(new_comment, request, *args,
                                          **kwargs)
            }

    def update_comment(self, request, review, comment, update_fields=(),
                       extra_fields={}, is_reply=False, **kwargs):
        """Update an existing comment based on the requested data.

        This will modify a comment, setting new fields requested by the caller.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            review (reviewboard.reviews.models.review.Review):
                The review owning the comment.

            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to update.

            update_fields (list of unicode, optional):
                The model fields that can be updated through the API.

            extra_fields (dict, optional):
                Extra fields from the request not otherwise handled by the
                API resource. Any ``extra_data`` modifications from this will
                be applied to the comment.

            is_reply (bool, optional):
                Whether this is a reply to another comment.

            **kwargs (dict):
                Keyword arguments representing additional fields handled by
                the API resource. Any that are also listed in ``fields`` will
                be set on the model.

        Returns:
            tuple or djblets.webapi.errors.WebAPIError:
            Either a successful payload containing the comment, or an error
            payload.
        """
        if is_reply:
            if not resources.review_reply.has_modify_permissions(request,
                                                                 review):
                return self.get_no_access_error(request)
        else:
            # Determine whether or not we're updating the issue status.
            if self.should_update_issue_status(comment, **kwargs):
                return self.update_issue_status(request, self, **kwargs)

            if not resources.review.has_modify_permissions(request, review):
                return self.get_no_access_error(request)

            # If we've updated the comment from having no issue opened,
            # to having an issue opened, we need to set the issue status
            # to OPEN.
            if not comment.issue_opened and kwargs.get('issue_opened', False):
                comment.issue_status = BaseComment.OPEN

            # If we've updated the comment from having an issue opened to
            # having no issue opened, set the issue status back to null.
            if comment.issue_opened and not kwargs.get('issue_opened', True):
                comment.issue_status = None

        for field in ('issue_opened',) + update_fields:
            value = kwargs.get(field, None)

            if value is not None:
                if isinstance(value, str):
                    value = value.strip()

                setattr(comment, field, value)

        self.set_text_fields(comment, 'text', **kwargs)

        if not is_reply:
            try:
                self.import_extra_data(comment, comment.extra_data,
                                       extra_fields)
            except ImportExtraDataError as e:
                return e.error_payload

        comment.save()

        return 200, {
            self.item_result_key: comment,
        }

    def update_issue_status(self, request, comment_resource, *args, **kwargs):
        """Updates the issue status for a comment.

        Handles all of the logic for updating an issue status.
        """
        try:
            review_request = \
                resources.review_request.get_object(request, *args, **kwargs)
            comment = comment_resource.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        # Check permissions to change the issue status
        if not comment.can_change_issue_status(request.user):
            return self.get_no_access_error(request)

        # We can only update the status of an issue if an issue has been
        # opened
        if not comment.issue_opened:
            raise PermissionDenied

        comment._review_request = review_request
        issue_status = \
            BaseComment.issue_string_to_status(kwargs.get('issue_status'))

        # If the issue requires verification, ensure that only people who are
        # authorized can close it.
        if (comment.require_verification and
            issue_status in (BaseComment.RESOLVED, BaseComment.DROPPED) and
            comment.issue_status in (BaseComment.OPEN,
                                     BaseComment.VERIFYING_RESOLVED,
                                     BaseComment.VERIFYING_DROPPED) and
            not comment.can_verify_issue_status(request.user)):
            return self.get_no_access_error(request)

        # We can only update the status of the issue
        comment.issue_status = issue_status
        comment.save(update_fields=['issue_status'])

        last_activity_time = \
            review_request.get_last_activity_info()['timestamp']

        return 200, {
            comment_resource.item_result_key: comment,
            'last_activity_time': last_activity_time.isoformat(),
        }

    def should_update_issue_status(self, comment, issue_status=None,
                                   issue_opened=None, **kwargs):
        """Returns True if the comment should have its issue status updated.

        Determines if a comment should have its issue status updated based
        on the current state of the comment, the review, and the arguments
        passed in the request.
        """
        if not issue_status:
            return False

        issue_status = BaseComment.issue_string_to_status(issue_status)

        return (comment.review.get().public and
                (comment.issue_opened or issue_opened) and
                issue_status != comment.issue_status)



"""Base class for diff comment resources."""

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.utils.timesince import timesince
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import webapi_request_fields
from djblets.webapi.fields import (IntFieldType,
                                   ResourceFieldType,
                                   StringFieldType)

from reviewboard.reviews.models import Comment
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.base_comment import BaseCommentResource


class BaseDiffCommentResource(BaseCommentResource):
    """Base class for diff comment resources.

    Provides common fields and functionality for all diff comment resources.
    The list of comments cannot be modified from this resource.
    """

    model = Comment
    name = 'diff_comment'
    fields = dict({
        'first_line': {
            'type': IntFieldType,
            'description': 'The line number that the comment starts at.',
        },
        'num_lines': {
            'type': IntFieldType,
            'description': 'The number of lines the comment spans.',
        },
        'filediff': {
            'type': ResourceFieldType,
            'resource': 'reviewboard.webapi.resources.filediff.'
                        'FileDiffResource',
            'description': 'The per-file diff that the comment was made on.',
        },
        'interfilediff': {
            'type': ResourceFieldType,
            'resource': 'reviewboard.webapi.resources.filediff.'
                        'FileDiffResource',
            'description': "The second per-file diff in an interdiff that "
                           "the comment was made on. This will be ``null`` if "
                           "the comment wasn't made on an interdiff.",
        },
    }, **BaseCommentResource.fields)

    uri_object_key = 'comment_id'

    allowed_methods = ('GET',)

    def get_queryset(self,
                     request,
                     review_request_id=None,
                     review_id=None,
                     is_list=False,
                     *args, **kwargs):
        """Return a queryset for Comment models.

        If the queryset is being used for a list of comment resources,
        then this can be further filtered by passing ``?interdiff-revision=``
        on the URL to match the given interdiff revision, and
        ``?line=`` to match comments on the given line number.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            review_request_id (int, optional):
                The review request ID used to filter the results. If set,
                only comments from the given review request that are public
                or owned by the requesting user will be included.

            review_id (int, optional):
                The review ID used to filter the results. If set,
                only comments from the given review that are public
                or owned by the requesting user will be included.

            is_list (bool, optional):
                Whether the incoming HTTP request is for list resources.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for Comment models.
        """
        q = Q()

        if review_request_id is not None:
            try:
                review_request = resources.review_request.get_object(
                    request, review_request_id=review_request_id,
                    *args, **kwargs)
            except ObjectDoesNotExist:
                raise self.model.DoesNotExist

            q &= Q(review__isnull=False)
            q &= Q(
                Q(filediff__diffset__history__review_request=review_request) |
                Q(filediff__diffset__review_request_draft__review_request=review_request))

        if is_list:
            if review_id:
                q &= Q(review=review_id)

            if 'interdiff-revision' in request.GET:
                interdiff_revision = int(request.GET['interdiff-revision'])
                q &= Q(interfilediff__diffset__revision=interdiff_revision)

            if 'line' in request.GET:
                q &= Q(first_line=int(request.GET['line']))

        queryset = self.model.objects.filter(q)
        order_by = kwargs.get('order-by', None)

        if order_by:
            queryset = queryset.order_by(*[
                field
                for field in order_by.split(',')
                if '__' not in field  # Don't allow joins
            ])

        return queryset

    def serialize_public_field(self, obj, **kwargs):
        return obj.review.get().public

    def serialize_timesince_field(self, obj, **kwargs):
        return timesince(obj.timestamp)

    def serialize_user_field(self, obj, **kwargs):
        return obj.review.get().user

    @webapi_request_fields(
        optional={
            'interdiff-revision': {
                'type': IntFieldType,
                'description': 'The second revision in an interdiff revision '
                               'range. The comments will be limited to this '
                               'range.',
            },
            'line': {
                'type': IntFieldType,
                'description': 'The line number that each comment must '
                               'start on.',
            },
            'order-by': {
                'type': StringFieldType,
                'description': 'Comma-separated list of fields to order by.',
                'added_in': '1.7.10',
            },
        },
        allow_unknown=True
    )
    @augment_method_from(BaseCommentResource)
    def get_list(self, request, review_id=None, *args, **kwargs):
        pass

    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Returns information on the comment."""
        pass



"""A comment on a file attachment."""

from __future__ import annotations

import logging
from typing import Optional, TYPE_CHECKING, cast

from django.db import models
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from typing_extensions import NotRequired, TypedDict

from reviewboard.attachments.models import FileAttachment
from reviewboard.reviews.managers import FileAttachmentCommentManager
from reviewboard.reviews.models.base_comment import BaseComment

if TYPE_CHECKING:
    from typing import ClassVar

    from django.utils.safestring import SafeText

    from reviewboard.diffviewer.models import DiffSet, FileDiff
    from reviewboard.reviews.ui.base import ReviewUI


logger = logging.getLogger(__name__)


class FileAttachmentCommentRevisionInfo(TypedDict):
    """Information about the diff revisions for a file attachment comment.

    For file attachment comments which are attached to files in a diff, we
    need to be able to get back the revision information.

    Version Added:
        7.0
    """

    #: The revision of the diff that the comment is on.
    diff_revision: int

    #: The PK of the DiffSet that the comment is on.
    diffset_id: int

    #: The interdiff revision, if present.
    interdiff_revision: Optional[int]

    #: The PK of the interdiff DiffSet, when present.
    interdiffset_id: Optional[int]

    #: The base commit ID, if present.
    base_commit_id: Optional[int]

    #: The tip commit ID, if present.
    tip_commit_id: Optional[int]

    #: The last commit ID in the commit series.
    #:
    #: This is used for some other computation when ``tip_commit_id`` is
    #: None.
    last_commit_id: NotRequired[int]

    #: The filediff for the modified version of the file.
    modified_filediff: FileDiff

    #: The diffset for the modified version of the file.
    modified_diffset: DiffSet


class FileAttachmentComment(BaseComment):
    """A comment on a file attachment."""

    anchor_prefix = 'fcomment'
    comment_type = 'file'

    file_attachment = models.ForeignKey(
        FileAttachment,
        on_delete=models.CASCADE,
        verbose_name=_('file attachment'),
        related_name='comments')
    diff_against_file_attachment = models.ForeignKey(
        FileAttachment,
        on_delete=models.CASCADE,
        verbose_name=_('diff against file attachment'),
        related_name='diffed_against_comments',
        null=True,
        blank=True)

    objects: ClassVar[FileAttachmentCommentManager] = \
        FileAttachmentCommentManager()

    @cached_property
    def review_ui(self) -> Optional[ReviewUI]:
        """Return a ReviewUI appropriate for this comment.

        If a ReviewUI is available for this type of file, an instance of
        one will be returned that's associated with this comment's
        FileAttachment and the one being diffed against (if any).

        Returns:
            reviewboard.reviews.ui.base.ReviewUI:
            The Review UI instance, if one is available.
        """
        review_ui = self.file_attachment.review_ui

        if not review_ui:
            return None

        if review_ui.supports_diffing and self.diff_against_file_attachment:
            review_ui.set_diff_against(self.diff_against_file_attachment)

        return review_ui

    @property
    def thumbnail(self) -> Optional[SafeText]:
        """Return the thumbnail for this comment, if any, as HTML.

        The thumbnail will be generated from the appropriate ReviewUI,
        if there is one for this type of file.

        Returns:
            django.utils.safestring.SafeText:
            The thumbnail for the file comment.
        """
        review_ui = self.review_ui

        if review_ui:
            try:
                return review_ui.get_comment_thumbnail(self)
            except Exception as e:
                logger.exception('Error when calling get_comment_thumbnail '
                                 'for ReviewUI %r: %s',
                                 review_ui, e)

        return mark_safe('')

    def get_absolute_url(self) -> Optional[str]:
        """Return the URL for this comment.

        Returns:
            str:
            The URL to link to for the comment.
        """
        review_ui = self.review_ui

        if review_ui:
            try:
                return review_ui.get_comment_link_url(self)
            except Exception as e:
                logger.exception('Error when calling get_comment_link_url '
                                 'for ReviewUI %r: %s',
                                 review_ui, e)

        return self.file_attachment.get_absolute_url()

    def get_link_text(self) -> Optional[str]:
        """Return the text for the link to the file.

        Returns:
            str:
            The text to use for the link to the file.
        """
        review_ui = self.review_ui

        if review_ui:
            try:
                return review_ui.get_comment_link_text(self)
            except Exception as e:
                logger.exception('Error when calling get_comment_link_text '
                                 'for ReviewUI %r: %s',
                                 review_ui, e)

        return self.file_attachment.display_name

    def attachment_is_public(self) -> bool:
        """Return whether the attachment(s) being commented on are public.

        Returns:
            bool:
            True if the file attachment (and diff against file attachment, if
            applicable) is public.
        """
        return (self.file_attachment.review_request.exists() or
                self.file_attachment.inactive_review_request.exists())

    def get_comment_diff_revision_info(
        self,
    ) -> Optional[FileAttachmentCommentRevisionInfo]:
        """Return the revision info for the comment.

        Version Added:
            7.0

        Returns:
            FileAttachmentCommentRevisionInfo:
            The revision information.
        """
        if not self.file_attachment.is_from_diff:
            return None

        diff_revision: int
        diffset_id: int
        interdiff_revision: Optional[int] = None
        interdiffset_id: Optional[int] = None

        modified_filediff = self.file_attachment.added_in_filediff
        modified_diffset = modified_filediff.diffset

        base_commit_id: Optional[int] = None
        tip_commit_id: Optional[int] = None

        diff_attachment = self.diff_against_file_attachment

        if diff_attachment is None:
            # This was a newly-added file.
            diff_revision = modified_diffset.revision
            diffset_id = modified_diffset.pk
            tip_commit_id = modified_filediff.commit_id
        else:
            orig_filediff = diff_attachment.added_in_filediff

            if orig_filediff is None:
                # The comment was made against a single diff revision.
                diff_revision = modified_diffset.revision
                diffset_id = modified_diffset.pk
                tip_commit_id = modified_filediff.commit_id
            else:
                orig_diffset = orig_filediff.diffset

                if orig_diffset.revision == modified_diffset.revision:
                    # The comment was made against a commit range on a single
                    # revision.
                    diff_revision = modified_diffset.revision
                    diffset_id = modified_diffset.pk
                    base_commit_id = orig_filediff.commit_id
                    tip_commit_id = modified_filediff.commit_id
                else:
                    # The comment was made on an interdiff.
                    diff_revision = orig_diffset.revision
                    diffset_id = orig_diffset.pk
                    interdiff_revision = modified_diffset.revision
                    interdiffset_id = modified_diffset.pk

        result: FileAttachmentCommentRevisionInfo = {
            'diff_revision': diff_revision,
            'diffset_id': diffset_id,
            'interdiff_revision': interdiff_revision,
            'interdiffset_id': interdiffset_id,
            'base_commit_id': base_commit_id,
            'tip_commit_id': tip_commit_id,
            'modified_filediff': modified_filediff,
            'modified_diffset': modified_diffset,
        }

        if tip_commit_id:
            last_commit_id = cast(int, (
                modified_diffset.commits
                .order_by('-pk')
                .values_list('pk', flat=True)
            )[0])

            # We don't specify the tip commit ID in the URL if it's the last
            # commit in the series.
            if tip_commit_id == last_commit_id:
                result['last_commit_id'] = last_commit_id
                result['tip_commit_id'] = None

        return result

    class Meta(BaseComment.Meta):
        """Metadata for the FileAttachmentComment model."""

        db_table = 'reviews_fileattachmentcomment'
        verbose_name = _('File Attachment Comment')
        verbose_name_plural = _('File Attachment Comments')



"""Testing utilities for building expected queries for Local Sites.

Version Added:
    5.0.7
"""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING, Union

from django.contrib.auth.models import User
from django.db.models import Q, Value

from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries
    from django.contrib.auth.models import AnonymousUser


def get_local_site_by_name_equeries(
    *,
    local_site: LocalSite,
) -> ExpectedQueries:
    """Return expected queries for accessing a Local Site by name.

    Version Added:
        5.0.7

    Args:
        local_site (reviewboard.site.models.LocalSite):
            The Local Site being accessed.

    Returns:
        list of dict:
        The list of expected queries.
    """
    return [
        {
            '__note__': f'Fetch the Local Site "{local_site.name}"',
            'model': LocalSite,
            'where': Q(name=local_site.name),
        },
    ]


def get_local_site_is_accessible_by_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: LocalSite,
) -> ExpectedQueries:
    """Return expected queries for Local Site accessibility checks.

    This corresponds to a call to :py:meth:`LocalSite.is_accessible_by()
    <reviewboard.site.models.LocalSite.is_accessible_by>`.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite):
            The Local Site being accessed.

    Returns:
        list of dict:
        The list of expected queries.
    """
    if user.is_superuser or user.is_anonymous:
        return []

    return [
        {
            '__note__': 'Check if the user is a member of the Local Site',
            'model': User,
            'annotations': {
                'a': Value(1),
            },
            'join_types': {
                'site_localsite_users': 'INNER JOIN',
            },
            'limit': 1,
            'num_joins': 1,
            'tables': {
                'auth_user',
                'site_localsite_users',
            },
            'where': (Q(local_site__id=local_site.pk) &
                      Q(pk=user.pk)),
        },
    ]


def get_local_site_is_mutable_by_equeries(
    *,
    user: User,
    local_site: LocalSite,
    note: Optional[str] = None,
) -> ExpectedQueries:
    """Return expected queries for Local Site mutability checks.

    This corresponds to a call to :py:meth:`LocalSite.is_mutable_by()
    <reviewboard.site.models.LocalSite.is_mutable_by>`.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite):
            The Local Site being accessed.

        note (str, optional):
            An optional note to use, instead of the default.

    Returns:
        list of dict:
        The list of expected queries.
    """
    if user.is_superuser or user.is_anonymous:
        return []

    return [
        {
            '__note__': note or (
                'Check if the user is an admin of the Local Site'
            ),
            'model': User,
            'annotations': {
                'a': Value(1),
            },
            'join_types': {
                'site_localsite_admins': 'INNER JOIN',
            },
            'limit': 1,
            'num_joins': 1,
            'tables': {
                'auth_user',
                'site_localsite_admins',
            },
            'where': (Q(local_site_admins__id=local_site.pk) &
                      Q(pk=user.pk)),
        },
    ]


def get_check_local_site_access_equeries(
    *,
    user: Union[AnonymousUser, User],
    local_site: LocalSite,
) -> ExpectedQueries:
    """Return expected queries for the @check_local_site_access decorator.

    This corresponds to a call to
    :py:func:`~reviewboard.site.decorators.check_local_site_access`.

    Version Added:
        5.0.7

    Args:
        user (django.contrib.auth.models.AnonymousUser or
              django.contrib.auth.models.User):
            The user that's checked for access.

        local_site (reviewboard.site.models.LocalSite):
            The Local Site being accessed.

    Returns:
        list of dict:
        The list of expected queries.
    """
    return (get_local_site_by_name_equeries(local_site=local_site) +
            get_local_site_is_accessible_by_equeries(user=user,
                                                     local_site=local_site))



from djblets.features.testing import override_feature_check
from djblets.testing.decorators import add_fixtures

from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import root_item_mimetype
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.urls import get_root_url


class ResourceTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the RootResource APIs."""
    fixtures = ['test_users']
    sample_api_url = '/'
    resource = resources.root
    test_http_methods = ('DELETE', 'PUT', 'POST')

    def setup_http_not_allowed_item_test(self, user):
        return get_root_url()

    def setup_http_not_allowed_list_test(self, user):
        return get_root_url()

    def test_get(self):
        """Testing the GET / API"""
        rsp = self.api_get(get_root_url(),
                           expected_mimetype=root_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('uri_templates', rsp)
        self.assertIn('repository', rsp['uri_templates'])
        self.assertEqual(rsp['uri_templates']['repository'],
                         'http://testserver/api/repositories/{repository_id}/')

        self._check_common_root_fields(rsp)

    @add_fixtures(['test_users', 'test_site'])
    def test_get_with_site(self):
        """Testing the GET / API with local sites"""
        self._login_user(local_site=True)
        rsp = self.api_get(get_root_url('local-site-1'),
                           expected_mimetype=root_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('uri_templates', rsp)
        self.assertIn('repository', rsp['uri_templates'])
        self.assertEqual(rsp['uri_templates']['repository'],
                         'http://testserver/s/local-site-1/api/'
                         'repositories/{repository_id}/')

        self._check_common_root_fields(rsp)

    @add_fixtures(['test_users', 'test_site'])
    def test_get_with_site_no_access(self):
        """Testing the GET / API without access to local site"""
        self.api_get(get_root_url('local-site-1'), expected_status=403)

    @add_fixtures(['test_users', 'test_site'])
    def test_get_with_site_and_cache(self):
        """Testing the GET / API with multiple local sites"""
        # djblets had a bug where the uri_templates were cached without any
        # consideration of the local site (or, more generally, the base uri).
        # In this case, fetching /s/<local_site>/api/ might return uri
        # templates for someone else's site. This was breaking rbt post.
        self.test_get_with_site()

        rsp = self.api_get(get_root_url('local-site-2'),
                           expected_mimetype=root_item_mimetype)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('uri_templates', rsp)
        self.assertIn('repository', rsp['uri_templates'])
        self.assertEqual(rsp['uri_templates']['repository'],
                         'http://testserver/s/local-site-2/api/'
                         'repositories/{repository_id}/')

    def test_get_capability_dvcs_enabled(self):
        """Testing the GET / API for capabilities with the DVCS feature enabled
        """
        with override_feature_check(dvcs_feature.feature_id, True):
            rsp = self.api_get(get_root_url(),
                               expected_mimetype=root_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('capabilities', rsp)

        caps = rsp['capabilities']
        self.assertIn('review_requests', caps)

        review_request_caps = caps['review_requests']
        self.assertIn('supports_history', review_request_caps)
        self.assertTrue(review_request_caps['supports_history'])

    def test_get_capability_dvcs_disabled(self):
        """Testing the GET / API for capabilities with the DVCS feature
        disabled
        """
        with override_feature_check(dvcs_feature.feature_id, False):
            self.assertFalse(dvcs_feature.is_enabled())
            rsp = self.api_get(get_root_url(),
                               expected_mimetype=root_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('capabilities', rsp)

        caps = rsp['capabilities']
        self.assertIn('review_requests', caps)
        self.assertNotIn('supports_history', caps['review_requests'])

    def test_get_uri_templates(self):
        """Testing that the RootResource has a URI templates dictionary that
        matches our expectations
        """
        expected_uri_templates = {
            'api_token':
                'http://testserver/api/users/{username}/api-tokens/'
                '{api_token_id}/',
            'api_tokens': 'http://testserver/api/users/{username}/api-tokens/',
            'archived_review_request':
                'http://testserver/api/users/{username}/'
                'archived-review-requests/{review_request_id}/',
            'archived_review_requests':
                'http://testserver/api/users/{username}/'
                'archived-review-requests/',
            'commit_validation': 'http://testserver/api/validation/commits/',
            'default_reviewer':
                'http://testserver/api/default-reviewers/'
                '{default_reviewer_id}/',
            'default_reviewers': 'http://testserver/api/default-reviewers/',
            'diff':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/',
            'diff_context':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diff-context/',
            'diff_validation': 'http://testserver/api/validation/diffs/',
            'diffs':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/',
            'extension': 'http://testserver/api/extensions/{extension_name}/',
            'extensions': 'http://testserver/api/extensions/',
            'group': 'http://testserver/api/groups/{group_name}/',
            'groups': 'http://testserver/api/groups/',
            'hosting_service':
                'http://testserver/api/hosting-services/{hosting_service_id}/',
            'hosting_service_account':
                'http://testserver/api/hosting-service-accounts/{account_id}/',
            'hosting_service_accounts':
                'http://testserver/api/hosting-service-accounts/',
            'hosting_services': 'http://testserver/api/hosting-services/',
            'info': 'http://testserver/api/info/',
            'muted_review_request':
                'http://testserver/api/users/{username}/muted-review-requests/'
                '{review_request_id}/',
            'muted_review_requests':
                'http://testserver/api/users/{username}/'
                'muted-review-requests/',
            'oauth_app': 'http://testserver/api/oauth-apps/{app_id}/',
            'oauth_apps': 'http://testserver/api/oauth-apps/',
            'oauth_token':
                'http://testserver/api/oauth-tokens/{oauth_token_id}/',
            'oauth_tokens': 'http://testserver/api/oauth-tokens/',
            'remote_repositories':
                'http://testserver/api/hosting-service-accounts/{account_id}/'
                'remote-repositories/',
            'remote_repository':
                'http://testserver/api/hosting-service-accounts/{account_id}/'
                'remote-repositories/{repository_id}/',
            'repositories': 'http://testserver/api/repositories/',
            'repository':
                'http://testserver/api/repositories/{repository_id}/',
            'review':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/',
            'review_draft':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/draft/',
            'review_group_user':
                'http://testserver/api/groups/{group_name}/users/{username}/',
            'review_group_users':
                'http://testserver/api/groups/{group_name}/users/',
            'review_request':
                'http://testserver/api/review-requests/{review_request_id}/',
            'review_requests': 'http://testserver/api/review-requests/',
            'reviews':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/',
            'root': 'http://testserver/api/',
            'screenshot':
                'http://testserver/api/review-requests/{review_request_id}/'
                'screenshots/{screenshot_id}/',
            'screenshot_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/screenshot-comments/{comment_id}/',
            'screenshot_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'screenshots/{screenshot_id}/screenshot-comments/',
            'screenshots':
                'http://testserver/api/review-requests/{review_request_id}/'
                'screenshots/',
            'search': 'http://testserver/api/search/{username}/',
            'session': 'http://testserver/api/session/',
            'user': 'http://testserver/api/users/{username}/',
            'user_file_attachment':
                'http://testserver/api/users/{username}/user-file-attachments/'
                '{file_attachment_id}/',
            'user_file_attachments':
                'http://testserver/api/users/{username}/'
                'user-file-attachments/',
            'users': 'http://testserver/api/users/',
            'validation': 'http://testserver/api/validation/',
            'watched': 'http://testserver/api/users/{username}/watched/',
            'watched_review_group':
                'http://testserver/api/users/{username}/watched/'
                'review-groups/{watched_obj_id}/',
            'watched_review_groups':
                'http://testserver/api/users/{username}/watched/'
                'review-groups/',
            'watched_review_request':
                'http://testserver/api/users/{username}/watched/'
                'review-requests/{watched_obj_id}/',
            'watched_review_requests':
                'http://testserver/api/users/{username}/watched/'
                'review-requests/',
            'webhook': 'http://testserver/api/webhooks/{webhook_id}/',
            'webhooks': 'http://testserver/api/webhooks/',

            # Added in 5.0.1.
            'all_general_comments': 'http://testserver/api/general-comments/',
            'all_reviews': 'http://testserver/api/reviews/',
            'file_diff_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/{filediff_id}/diff-comments/',
            'review_reply_screenshot_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/screenshot-comments/'
                '{comment_id}/',
            'review_reply_screenshot_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/screenshot-comments/',

            # Added in 5.0.2.
            'all_diff_comments': 'http://testserver/api/diff-comments/',
            'all_file_attachment_comments':
                'http://testserver/api/file-attachment-comments/',
            'diff_commit':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/commits/{commit_id}/',
            'diff_commits':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/commits/',
            'file_diff':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/{filediff_id}/',
            'file_diffs':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/',
            'file_diff_original_file':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/{filediff_id}/original-file/',
            'file_diff_patched_file':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/{filediff_id}/patched-file/',
            'repository_branches':
                'http://testserver/api/repositories/{repository_id}/branches/',
            'repository_commits':
                'http://testserver/api/repositories/{repository_id}/commits/',
            'repository_group':
                'http://testserver/api/repositories/{repository_id}/groups/'
                '{group_name}/',
            'repository_groups':
                'http://testserver/api/repositories/{repository_id}/groups/',
            'repository_info':
                'http://testserver/api/repositories/{repository_id}/info/',
            'repository_user':
                'http://testserver/api/repositories/{repository_id}/users/'
                '{username}/',
            'repository_users':
                'http://testserver/api/repositories/{repository_id}/users/',
            'review_diff_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/diff-comments/{comment_id}/',
            'review_diff_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/diff-comments/',
            'review_general_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/general-comments/{comment_id}/',
            'review_general_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/general-comments/',
            'review_replies':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/',
            'review_reply':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/',
            'review_reply_draft':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/draft/',
            'review_reply_diff_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/diff-comments/'
                '{comment_id}/',
            'review_reply_diff_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/diff-comments/',
            'review_reply_file_attachment_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/'
                'file-attachment-comments/{comment_id}/',
            'review_reply_file_attachment_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/'
                'file-attachment-comments/',
            'review_reply_general_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/general-comments/'
                '{comment_id}/',
            'review_reply_general_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/general-comments/',
            'review_request_change':
                'http://testserver/api/review-requests/{review_request_id}/'
                'changes/{change_id}/',
            'review_request_changes':
                'http://testserver/api/review-requests/{review_request_id}/'
                'changes/',
            'review_request_draft':
                'http://testserver/api/review-requests/{review_request_id}/'
                'draft/',
            'review_request_file_attachment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'file-attachments/{file_attachment_id}/',
            'review_request_file_attachments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'file-attachments/',
            'review_request_file_attachment_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'file-attachments/{file_attachment_id}/'
                'file-attachment-comments/',
            'review_request_last_update':
                'http://testserver/api/review-requests/{review_request_id}/'
                'last-update/',
            'review_request_status_update':
                'http://testserver/api/review-requests/{review_request_id}/'
                'status-updates/{status_update_id}/',
            'review_request_status_updates':
                'http://testserver/api/review-requests/{review_request_id}/'
                'status-updates/',

            # Deprecated in 5.0.2.
            'branches':
                'http://testserver/api/repositories/{repository_id}/branches/',
            'change':
                'http://testserver/api/review-requests/{review_request_id}/'
                'changes/{change_id}/',
            'changes':
                'http://testserver/api/review-requests/{review_request_id}/'
                'changes/',
            'commit':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/commits/{commit_id}/',
            'commits':
                'http://testserver/api/repositories/{repository_id}/commits/',
            'diff_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/diff-comments/'
                '{comment_id}/',
            'diff_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/diff-comments/',
            'diff_file_attachment':
                'http://testserver/api/repositories/{repository_id}/'
                'diff-file-attachments/{file_attachment_id}/',
            'diff_file_attachments':
                'http://testserver/api/repositories/{repository_id}/'
                'diff-file-attachments/',
            'draft':
                'http://testserver/api/review-requests/{review_request_id}/'
                'draft/',
            'file':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/{filediff_id}/',
            'files':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/',
            'file_attachment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'file-attachments/{file_attachment_id}/',
            'file_attachments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'file-attachments/',
            'file_attachment_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/file-attachment-comments/{comment_id}/',
            'file_attachment_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/file-attachment-comments/',
            'general_comment':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/general-comments/{comment_id}/',
            'general_comments':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/general-comments/',
            'last_update':
                'http://testserver/api/review-requests/{review_request_id}/'
                'last-update/',
            'original_file':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/{filediff_id}/original-file/',
            'patched_file':
                'http://testserver/api/review-requests/{review_request_id}/'
                'diffs/{diff_revision}/files/{filediff_id}/patched-file/',
            'replies':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/',
            'reply':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/{reply_id}/',
            'reply_draft':
                'http://testserver/api/review-requests/{review_request_id}/'
                'reviews/{review_id}/replies/draft/',
            'status_update':
                'http://testserver/api/review-requests/{review_request_id}/'
                'status-updates/{status_update_id}/',
            'status_updates':
                'http://testserver/api/review-requests/{review_request_id}/'
                'status-updates/',
        }

        rsp = self.api_get(get_root_url(),
                           expected_mimetype=root_item_mimetype)
        current_uri_templates = rsp['uri_templates']

        self.assertDictEqual(current_uri_templates,
                             expected_uri_templates)

    def _check_common_root_fields(self, item_rsp):
        self.assertIn('product', item_rsp)
        self.assertIn('site', item_rsp)
        self.assertIn('capabilities', item_rsp)

        caps = item_rsp['capabilities']
        self.assertIn('diffs', caps)

        diffs_caps = caps['diffs']
        self.assertTrue(diffs_caps['moved_files'])
        self.assertTrue(diffs_caps['base_commit_ids'])

        diff_validation_caps = diffs_caps['validation']
        self.assertTrue(diff_validation_caps['base_commit_ids'])

        review_request_caps = caps['review_requests']
        self.assertTrue(review_request_caps['commit_ids'])

        text_caps = caps['text']
        self.assertTrue(text_caps['markdown'])



"""Mixins for notifications tests.

Version Added:
    6.0
"""

from typing import List, TYPE_CHECKING

from django.contrib.auth.models import User
from django.core import mail
from djblets.mail.utils import build_email_address_for_user
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin.siteconfig import load_site_config, settings_map
from reviewboard.notifications.email.utils import get_email_addresses_for_group
from reviewboard.reviews.models import Group
from reviewboard.testing import TestCase


if TYPE_CHECKING:
    class HelperBase(TestCase):
        pass
else:
    class HelperBase:
        pass


class EmailTestHelper(HelperBase):
    """A mixin for test cases which need to verify e-mail sending."""

    email_siteconfig_settings = {}

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        mail.outbox = []
        self.sender = 'noreply@example.com'
        self._old_email_settings = {}

        if self.email_siteconfig_settings:
            siteconfig = SiteConfiguration.objects.get_current()
            needs_reload = False

            for key, value in self.email_siteconfig_settings.items():
                old_value = siteconfig.get(key)

                if old_value != value:
                    self._old_email_settings[key] = old_value
                    siteconfig.set(key, value)

                    if key in settings_map:
                        needs_reload = True

            if self._old_email_settings:
                siteconfig.save()

                if needs_reload:
                    load_site_config()

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()

        if self._old_email_settings:
            siteconfig = SiteConfiguration.objects.get_current()
            needs_reload = False

            for key, value in self._old_email_settings.items():
                self._old_email_settings[key] = siteconfig.get(key)
                siteconfig.set(key, value)

                if key in settings_map:
                    needs_reload = True

            siteconfig.save()

            if needs_reload:
                load_site_config()

    def assertValidRecipients(
        self,
        user_list: List[str],
        group_list: List[str] = [],
    ) -> None:
        """Assert that the e-mail recipient list is as expected.

        Args:
            user_list (list of str):
                The list of usernames for users who should be included on the
                email.

            group_list (list of str, optional):
                The list of group names for groups who should be included on
                the email.
        """
        recipient_list = mail.outbox[0].to + mail.outbox[0].cc
        self.assertEqual(len(recipient_list), len(user_list) + len(group_list))

        for user in user_list:
            self.assertTrue(build_email_address_for_user(
                User.objects.get(username=user)) in recipient_list,
                'user %s was not found in the recipient list' % user)

        groups = Group.objects.filter(name__in=group_list, local_site=None)

        for group in groups:
            for address in get_email_addresses_for_group(group):
                self.assertTrue(
                    address in recipient_list,
                    'group %s was not found in the recipient list' % address)



"""Unit tests for reviewboard.webapi.resources.validate_diffcommit."""

import base64
import json

from django.core.files.uploadedfile import SimpleUploadedFile
from djblets.features.testing import override_feature_checks
from djblets.webapi.errors import INVALID_ATTRIBUTE
from djblets.webapi.testing.decorators import webapi_test_template
from kgb import SpyAgency

from reviewboard.diffviewer.commit_utils import serialize_validation_info
from reviewboard.scmtools.models import Repository
from reviewboard.webapi.errors import (DIFF_EMPTY, DIFF_PARSE_ERROR,
                                       DIFF_TOO_BIG, INVALID_REPOSITORY)
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import validate_diffcommit_mimetype
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.mixins_ssl import SSLTestsMixin
from reviewboard.webapi.tests.urls import get_validate_diffcommit_url


class ResourceTests(SpyAgency, SSLTestsMixin, BaseWebAPITestCase,
                    metaclass=BasicTestsMetaclass):
    """Testing ValidateDiffCommitResource API."""

    resource = resources.validate_diffcommit
    sample_api_url = 'validation/validate-commit/'

    fixtures = ['test_scmtools', 'test_users']

    # The basic GET request *does* return JSON, but it is a singleton resource
    # whose item_result_key is only present in successful POST responses.
    basic_get_returns_json = False
    basic_post_success_status = 200

    def compare_item(self, item_rsp, item):
        """Compare a response to the item.

        This is intentionally a no-op because there will be no created model to
        compare against.

        Args:
            item_rsp (dict):
                The serialized response.

            item (object):
                The item to compare to. This is always ``None``.
        """
        pass

    def setup_http_not_allowed_item_test(self, user):
        return get_validate_diffcommit_url()

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        return (get_validate_diffcommit_url(local_site_name=local_site_name),
                validate_diffcommit_mimetype,
                None)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        repository = self.create_repository(tool_name='Git',
                                            with_local_site=with_local_site)

        post_data = {}

        if post_valid_data:
            self.spy_on(Repository.get_file_exists,
                        owner=Repository,
                        call_fake=lambda *args, **kwargs: True)

            validation_info = serialize_validation_info({
                'r1': {
                    'parent_id': 'r0',
                    'tree': {
                        'added': [{
                            'filename': 'README',
                            'revision': '94bdd3e',
                        }],
                        'modified': [],
                        'removed': [],
                    },
                },
            })
            diff = SimpleUploadedFile('diff',
                                      self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                      content_type='text/x-patch')
            post_data = {
                'commit_id': 'r2',
                'parent_id': 'r1',
                'diff': diff,
                'validation_info': validation_info,
                'repository': repository.pk,
            }

        return (get_validate_diffcommit_url(local_site_name=local_site_name),
                validate_diffcommit_mimetype,
                post_data,
                [])

    def check_post_result(self, user, rsp):
        self.assertIn('commit_validation', rsp)
        self.assertIn('validation_info', rsp['commit_validation'])

    @webapi_test_template
    def test_post_diff_too_big(self):
        """Testing the POST <URL> API with a diff that is too big"""
        repo = self.create_repository(tool_name='Git')

        with self.siteconfig_settings({'diffviewer_max_diff_size': 1},
                                      reload_settings=False):
            with override_feature_checks(self.override_features):
                rsp = self.api_post(
                    get_validate_diffcommit_url(),
                    {

                        'commit_id': 'r1',
                        'parent_id': 'r0',
                        'diff': SimpleUploadedFile(
                            'diff',
                            self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                            content_type='text/x-patch'),
                        'repository': repo.name,
                    },
                    expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DIFF_TOO_BIG.code)
        self.assertEqual(rsp['max_size'], 1)

    @webapi_test_template
    def test_post_diff_empty(self):
        """Testing the POST <URL> API with an empty diff"""
        repo = self.create_repository(tool_name='Git')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_validate_diffcommit_url(),
                {

                    'commit_id': 'r1',
                    'parent_id': 'r0',
                    'diff': SimpleUploadedFile('diff',
                                               b'    ',
                                               content_type='text/x-patch'),
                    'repository': repo.name,
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DIFF_EMPTY.code)

    @webapi_test_template
    def test_post_diff_parser_error(self):
        """Testing the POST <URL> API with a diff that does not parse"""
        repo = self.create_repository(tool_name='Git')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_validate_diffcommit_url(),
                {

                    'commit_id': 'r1',
                    'parent_id': 'r0',
                    'diff': SimpleUploadedFile('diff',
                                               b'not a valid diff at all.',
                                               content_type='text/x-patch'),
                    'repository': repo.name,
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DIFF_PARSE_ERROR.code)
        self.assertEqual(rsp['linenum'], 0)

    @webapi_test_template
    def test_post_repo_no_history_support(self):
        """Testing the POST <URL> API with a repository that does not support
        history
        """
        repo = self.create_repository(tool_name='Test')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_validate_diffcommit_url(),
                {

                    'commit_id': 'r1',
                    'parent_id': 'r0',
                    'diff': SimpleUploadedFile(
                        'diff',
                        self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                        content_type='text/x-patch'),
                    'repository': repo.name,
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_ATTRIBUTE.code)
        self.assertEqual(
            rsp['reason'],
            'The "%s" repository does not support review requests created '
            'with history.'
            % repo.name)

    @webapi_test_template
    def test_post_repo_does_not_exist(self):
        """Testing the POST <URL> API with a repository that does not exist"""
        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_validate_diffcommit_url(),
                {
                    'commit_id': 'r1',
                    'parent_id': 'r0',
                    'diff': SimpleUploadedFile(
                        'diff',
                        self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                        content_type='text/x-patch'),
                    'repository': 'nope',
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_REPOSITORY.code)

    @webapi_test_template
    def test_post_repo_no_access(self):
        """Testing the POST <URL> API with a repository the user does not have
        access to
        """
        repo = self.create_repository(public=False)

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_validate_diffcommit_url(),
                {
                    'commit_id': 'r1',
                    'parent_id': 'r0',
                    'diff': SimpleUploadedFile(
                        'diff',
                        self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                        content_type='text/x-patch'),
                    'repository': repo.name,
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_REPOSITORY.code)

    @webapi_test_template
    def test_post_with_ssl_error(self) -> None:
        """Testing the POST <URL> API with CertificateVerificationError"""
        repository = self.create_repository()

        self.run_ssl_cert_test(
            spy_func=repository.scmtool_class.file_exists,
            spy_owner=repository.scmtool_class,
            url=get_validate_diffcommit_url(),
            data={
                'commit_id': 'r1',
                'parent_id': 'r0',
                'diff': SimpleUploadedFile(
                    'diff',
                    self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                    content_type='text/x-patch'),
                'repository': repository.name,
            })

    @webapi_test_template
    def test_post_repo_multiple(self):
        """Testing the POST <URL> API with multiple matching repositories"""
        repo1 = self.create_repository(name='repo1')
        self.create_repository(name='repo2',
                               path=repo1.path)

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_validate_diffcommit_url(),
                {
                    'commit_id': 'r1',
                    'parent_id': 'r0',
                    'diff': SimpleUploadedFile(
                        'diff',
                        self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                        content_type='text/x-patch'),
                    'repository': repo1.path,
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_REPOSITORY.code)
        self.assertEqual(rsp['err']['msg'],
                         'Too many repositories matched "%s". Try specifying '
                         'the repository by name instead.'
                         % repo1.path)

    @webapi_test_template
    def test_post_parent_diff(self):
        """Testing the POST <URL> API with a parent diff"""
        def _exists(repository, path, revision, *args, **kwargs):
            return path == 'README' and revision == '94bdd3e'

        repo = self.create_repository(name='repo')
        self.spy_on(Repository.get_file_exists,
                    owner=Repository,
                    call_fake=_exists)

        parent_diff_contents = (
            b'diff --git a/README b/README\n'
            b'index 94bdd3e..f00f00 100644\n'
            b'--- README\n'
            b'+++ README\n'
            b'@@ -2 +2 @@\n'
            b'-blah blah\n'
            b'+foo bar\n'
        )
        parent_diff = SimpleUploadedFile('parent_diff', parent_diff_contents,
                                         content_type='text/x-patch')

        diff_contents = (
            b'diff --git a/README b/README\n'
            b'index f00f00..197009f 100644\n'
            b'--- README\n'
            b'+++ README\n'
            b'@@ -2 +2 @@\n'
            b'-foo bar\n'
            b'+blah!\n'
        )
        diff = SimpleUploadedFile('diff', diff_contents,
                                  content_type='text/x-patch')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_validate_diffcommit_url(),
                {
                    'commit_id': 'r1',
                    'parent_id': 'r0',
                    'diff': diff,
                    'parent_diff': parent_diff,
                    'repository': repo.name,
                },
                expected_mimetype=validate_diffcommit_mimetype,
                expected_status=200)

        self.assertEqual(rsp['stat'], 'ok')

        validation_info = json.loads(base64.b64decode(
            rsp['commit_validation']['validation_info']).decode('utf-8'))
        self.assertEqual(validation_info, {
            'r1': {
                'parent_id': 'r0',
                'tree': {
                    'added': [],
                    'modified': [
                        {
                            'filename': 'README',
                            'revision': '197009f',
                        },
                    ],
                    'removed': [],
                },
            },
        })

    @webapi_test_template
    def test_post_added_in_parent(self):
        """Testing the POST <URL> API with a subsequent commit that contains a
        file added in the parent diff
        """
        def _exists(repository, path, revision, *args, **kwargs):
            return path == 'README' and revision == '94bdd3e'

        initial_validation_info = {
            'r1': {
                'parent_id': 'r0',
                'tree': {
                    'added': [],
                    'modified': [],
                    'removed': [],
                },
            },
        }

        repo = self.create_repository(name='repo')

        self.spy_on(Repository.get_file_exists,
                    owner=Repository,
                    call_fake=_exists)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')

        with override_feature_checks(self.override_features):
            rsp = self.api_post(
                get_validate_diffcommit_url(),
                {
                    'commit_id': 'r2',
                    'parent_id': 'r1',
                    'diff': diff,
                    'repository': repo.name,
                    'validation_info': serialize_validation_info(
                        initial_validation_info),
                },
                expected_mimetype=validate_diffcommit_mimetype,
                expected_status=200)

        self.assertEqual(rsp['stat'], 'ok')



"""Unit tests for the DraftDiffResource."""

from __future__ import annotations

import base64

import kgb
from django.core.files.uploadedfile import SimpleUploadedFile
from djblets.features.testing import override_feature_check
from djblets.webapi.errors import (INVALID_ATTRIBUTE,
                                   INVALID_FORM_DATA,
                                   PERMISSION_DENIED)
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.diffviewer.commit_utils import (serialize_validation_info,
                                                 update_validation_info)
from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.diffviewer.models import DiffSet, FileDiff
from reviewboard.reviews.models import DefaultReviewer
from reviewboard.reviews.signals import review_request_diffset_uploaded
from reviewboard.scmtools.models import Repository
from reviewboard.webapi.errors import DIFF_TOO_BIG
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (diff_item_mimetype,
                                                diff_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.mixins_extra_data import (ExtraDataItemMixin,
                                                        ExtraDataListMixin)
from reviewboard.webapi.tests.urls import (get_draft_diff_item_url,
                                           get_draft_diff_list_url)


class ResourceListTests(kgb.SpyAgency, ExtraDataListMixin, BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Testing the DraftDiffResource list APIs."""
    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/draft/diffs/'
    resource = resources.draft_diff

    def compare_item(self, item_rsp, diffset):
        self.assertEqual(item_rsp['id'], diffset.pk)
        self.assertEqual(item_rsp['name'], diffset.name)
        self.assertEqual(item_rsp['revision'], diffset.revision)
        self.assertEqual(item_rsp['basedir'], diffset.basedir)
        self.assertEqual(item_rsp['base_commit_id'], diffset.base_commit_id)
        self.assertEqual(item_rsp['extra_data'], diffset.extra_data)

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user,
            publish=True)

        return get_draft_diff_list_url(review_request)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user,
            publish=True)

        if populate_items:
            items = [self.create_diffset(review_request, draft=True)]
        else:
            items = []

        return (get_draft_diff_list_url(review_request, local_site_name),
                diff_list_mimetype,
                items)

    def test_get_not_owner(self):
        """Testing the GET review-requests/<id>/draft/diffs/ API
        without owner with Permission Denied error
        """
        review_request = self.create_review_request(create_repository=True)
        self.assertNotEqual(review_request.submitter, self.user)
        self.create_diffset(review_request, draft=True)

        self.api_get(get_draft_diff_list_url(review_request),
                     expected_status=403)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        repository = self.create_repository(tool_name='Test')
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            repository=repository,
            submitter=user)

        if post_valid_data:
            diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                      content_type='text/x-patch')
            post_data = {
                'path': diff,
                'basedir': '/trunk',
                'base_commit_id': '1234',
            }
        else:
            post_data = {}

        return (get_draft_diff_list_url(review_request, local_site_name),
                diff_item_mimetype,
                post_data,
                [review_request])

    def check_post_result(self, user, rsp, review_request):
        self.assertIn('diff', rsp)
        item_rsp = rsp['diff']

        draft = review_request.get_draft()
        self.assertIsNotNone(draft)

        diffset = DiffSet.objects.get(pk=item_rsp['id'])
        self.assertEqual(diffset, draft.diffset)
        self.compare_item(item_rsp, diffset)

    def test_post_with_missing_data(self):
        """Testing the POST review-requests/<id>/draft/diffs/ API
        with Invalid Form Data
        """
        repository = self.create_repository(tool_name='Test')
        review_request = self.create_review_request(repository=repository,
                                                    submitter=self.user)

        rsp = self.api_post(get_draft_diff_list_url(review_request),
                            expected_status=400)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertIn('path', rsp['fields'])

        # Now test with a valid path and an invalid basedir.
        # This is necessary because basedir is "optional" as defined by
        # the resource, but may be required by the form that processes the
        # diff.
        review_request = self.create_review_request(repository=repository,
                                                    submitter=self.user)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_draft_diff_list_url(review_request),
            {'path': diff},
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertIn('basedir', rsp['fields'])

    def test_post_diffs_too_big(self):
        """Testing the POST review-requests/<id>/draft/diffs/ API
        with diff exceeding max size
        """
        repository = self.create_repository()
        review_request = self.create_review_request(repository=repository,
                                                    submitter=self.user)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        with self.siteconfig_settings({'diffviewer_max_diff_size': 2},
                                      reload_settings=False):
            rsp = self.api_post(
                get_draft_diff_list_url(review_request),
                {
                    'path': diff,
                    'basedir': "/trunk",
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DIFF_TOO_BIG.code)
        self.assertIn('reason', rsp)
        self.assertIn('max_size', rsp)
        self.assertEqual(rsp['max_size'], 2)

    @webapi_test_template
    def test_post_diff_with_history(self):
        """Testing the POST <URL> API with a diff and a review request created
        with history support
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True,
                                                    create_with_history=True)

        with override_feature_check(dvcs_feature.feature_id, True):
            rsp = self.api_post(
                get_draft_diff_list_url(review_request),
                {
                    'path': SimpleUploadedFile(
                        'diff', self.DEFAULT_GIT_FILEDIFF_DATA_DIFF),
                    'basedir': '',
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(
            rsp['reason'],
            'This review request was created with support for multiple '
            'commits.\n\n'
            'Create an empty diff revision and upload commits to that '
            'instead.')

    @webapi_test_template
    def test_post_empty_with_history(self):
        """Testing the POST <URL> API creates an empty DiffSet for a review
        request created with history support with the DVCS feature enabled
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True,
                                                    create_with_history=True)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            rsp = self.api_post(get_draft_diff_list_url(review_request), {},
                                expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['diff']

        diff = DiffSet.objects.get(pk=item_rsp['id'])
        self.compare_item(item_rsp, diff)
        self.assertEqual(diff.files.count(), 0)

    @webapi_test_template
    def test_post_empty_dvcs_disabled(self):
        """Testing the POST <URL> API without a diff with the DVCS feature
        disabled
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True,
                                                    create_with_history=False)

        with override_feature_check(dvcs_feature.feature_id, enabled=False):
            rsp = self.api_post(get_draft_diff_list_url(review_request), {},
                                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(rsp['fields'], {
            'path': ['This field is required.'],
        })

    @webapi_test_template
    def test_post_adds_default_reviewers(self):
        """Testing the POST <URL> API adds default reviewers"""
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True)
        group = self.create_review_group(name='group1')

        default_reviewer = DefaultReviewer.objects.create(
            name='default1',
            file_regex='.')
        default_reviewer.groups.add(group)
        default_reviewer.repository.add(review_request.repository)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_draft_diff_list_url(review_request),
            {
                'path': diff,
                'basedir': '/trunk',
                'base_commit_id': '1234',
            },
            expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        draft = review_request.get_draft()
        self.assertEqual(list(draft.target_groups.all()), [group])

    @webapi_test_template
    def test_post_emits_review_request_diffset_uploaded(self) -> None:
        """Testing the POST <URL> API emits a review_request_diffset_uploaded
        signal
        """
        def on_diffset_uploaded(diffset, review_request_draft, **kwargs):
            pass

        review_request_diffset_uploaded.connect(on_diffset_uploaded)
        self.spy_on(on_diffset_uploaded)

        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True)

        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_draft_diff_list_url(review_request),
            {
                'path': diff,
                'basedir': '/trunk',
                'base_commit_id': '1234',
            },
            expected_mimetype=diff_item_mimetype)

        draft = review_request.get_draft()
        diffset = DiffSet.objects.get(pk=rsp['diff']['id'])

        self.assertEqual(rsp['stat'], 'ok')
        self.assertSpyCalledWith(
            on_diffset_uploaded,
            diffset=diffset,
            review_request_draft=draft)

        review_request_diffset_uploaded.disconnect(on_diffset_uploaded)

    @webapi_test_template
    def test_post_with_history_review_request_diffset_uploaded(self) -> None:
        """Testing the POST <URL> API does not emit a
        review_request_diffset_uploaded signal for a review request created
        with history support
        """
        def on_diffset_uploaded(diffset, review_request_draft, **kwargs):
            pass

        review_request_diffset_uploaded.connect(on_diffset_uploaded)
        self.spy_on(on_diffset_uploaded)

        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True,
                                                    create_with_history=True)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            rsp = self.api_post(get_draft_diff_list_url(review_request), {},
                                expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertSpyNotCalled(on_diffset_uploaded)

        review_request_diffset_uploaded.disconnect(on_diffset_uploaded)


class ResourceItemTests(kgb.SpyAgency, ExtraDataItemMixin, BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Testing the DraftDiffResource item APIs."""
    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/draft/diffs/<revision>/'
    resource = resources.draft_diff

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        return get_draft_diff_item_url(review_request, 1)

    def setup_http_not_allowed_list_test(self, user):
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        return get_draft_diff_item_url(review_request, 1)

    def compare_item(self, item_rsp, diffset):
        self.assertEqual(item_rsp['id'], diffset.pk)
        self.assertEqual(item_rsp['name'], diffset.name)
        self.assertEqual(item_rsp['revision'], diffset.revision)
        self.assertEqual(item_rsp['basedir'], diffset.basedir)
        self.assertEqual(item_rsp['base_commit_id'], diffset.base_commit_id)
        self.assertEqual(item_rsp['extra_data'], diffset.extra_data)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user)
        diffset = self.create_diffset(review_request, draft=True)

        return (get_draft_diff_item_url(review_request, diffset.revision,
                                        local_site_name),
                diff_item_mimetype,
                diffset)

    def test_get_not_modified(self):
        """Testing the GET review-requests/<id>/draft/diffs/<revision>/ API
        with Not Modified response
        """
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=self.user)
        diffset = self.create_diffset(review_request, draft=True)

        self._testHttpCaching(
            get_draft_diff_item_url(review_request, diffset.revision),
            check_etags=True)

    def test_get_not_owner(self):
        """Testing the GET review-requests/<id>/draft/diffs/<revision>/ API
        without owner with Permission Denied error
        """
        review_request = self.create_review_request(create_repository=True)
        self.assertNotEqual(review_request.submitter, self.user)
        diffset = self.create_diffset(review_request, draft=True)

        self.api_get(
            get_draft_diff_item_url(review_request, diffset.revision),
            expected_status=403)

    @webapi_test_template
    def test_get_patch_and_inaccessible(self) -> None:
        """Testing the GET <URL> API with Accept: text/x-patch and
        inaccessible diff
        """
        repository = self.create_repository(public=False)

        review_request = self.create_review_request(repository=repository)
        self.assertNotEqual(review_request.submitter, self.user)

        diffset = self.create_diffset(review_request, draft=True)
        self.create_filediff(diffset)

        rsp = self.api_get(
            get_draft_diff_item_url(review_request, diffset.revision),
            HTTP_ACCEPT='text/x-patch',
            expected_status=403)

        self.assertEqual(
            rsp,
            {
                'err': {
                    'code': PERMISSION_DENIED.code,
                    'msg': PERMISSION_DENIED.msg,
                    'type': 'resource-permission-denied',
                },
                'stat': 'fail',
            })

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user)
        diffset = self.create_diffset(review_request, draft=True)

        return (get_draft_diff_item_url(review_request, diffset.revision,
                                        local_site_name),
                diff_item_mimetype,
                {},
                diffset,
                [])

    def check_put_result(self, user, item_rsp, diffset):
        diffset = DiffSet.objects.get(pk=diffset.pk)
        self.compare_item(item_rsp, diffset)

    @webapi_test_template
    def test_put_finalize(self):
        """Testing the PUT <URL> API with finalize_commit_series=1"""
        @self.spy_for(Repository.get_file_exists,
                      owner=Repository)
        def _get_file_exists(repository, path, revision, *args, **kwargs):
            self.assertEqual(path, 'README')
            self.assertEqual(revision, '94bdd3e')

            return True

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            commit = self.create_diffcommit(diffset=diffset)

            filediff = FileDiff.objects.get()

            cumulative_diff = SimpleUploadedFile('diff', filediff.diff,
                                                 content_type='text/x-patch')

            validation_info = update_validation_info(
                {},
                commit_id=commit.commit_id,
                parent_id=commit.parent_id,
                filediffs=[filediff])

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info(
                        validation_info),
                },
                expected_mimetype=diff_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.compare_item(rsp['diff'], diffset)

    @webapi_test_template
    def test_put_finalized_with_parent(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 and a parent
        diff
        """
        @self.spy_for(Repository.get_file_exists,
                      owner=Repository)
        def _get_file_exists(repository, path, revision, *args, **kwargs):
            self.assertEqual(path, 'README')
            self.assertEqual(revision, 'f00f00')

            return True

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            commit = self.create_diffcommit(diffset=diffset)

            filediff = FileDiff.objects.get()
            filediff.parent_diff = (
                b'diff --git a/README b/README\n'
                b'index f00f00..94bdd3e\n'
                b'--- a/README\n'
                b'+++ b/README\n'
            )
            filediff.extra_data[FileDiff._IS_PARENT_EMPTY_KEY] = True
            filediff.save(update_fields=('extra_data',))

            cumulative_diff = SimpleUploadedFile('diff', filediff.diff,
                                                 content_type='text/x-patch')

            parent_diff = SimpleUploadedFile('parent_diff',
                                             filediff.parent_diff,
                                             content_type='text/x-patch')

            validation_info = update_validation_info(
                {},
                commit_id=commit.commit_id,
                parent_id=commit.parent_id,
                filediffs=[filediff])

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'parent_diff': parent_diff,
                    'validation_info': serialize_validation_info(
                        validation_info),
                },
                expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.compare_item(rsp['diff'], diffset)

    @webapi_test_template
    def test_put_finalize_adds_default_reviewers(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 adds
        default reviewers
        """
        @self.spy_for(Repository.get_file_exists,
                      owner=Repository)
        def _get_file_exists(repository, path, revision, *args, **kwargs):
            self.assertEqual(path, 'README')
            self.assertEqual(revision, '94bdd3e')

            return True

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)

            # Create the state needed for the default reviewer.
            group = self.create_review_group(name='group1')

            default_reviewer = DefaultReviewer.objects.create(
                name='default1',
                file_regex='.')
            default_reviewer.groups.add(group)
            default_reviewer.repository.add(review_request.repository)

            # Create the state needed for the diff to post.
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            commit = self.create_diffcommit(diffset=diffset)

            filediff = FileDiff.objects.get()

            cumulative_diff = SimpleUploadedFile('diff', filediff.diff,
                                                 content_type='text/x-patch')

            validation_info = update_validation_info(
                {},
                commit_id=commit.commit_id,
                parent_id=commit.parent_id,
                filediffs=[filediff])

            # Post the diff.
            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info(
                        validation_info),
                },
                expected_mimetype=diff_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.compare_item(rsp['diff'], diffset)

            draft = review_request.get_draft()
            self.assertEqual(list(draft.target_groups.all()), [group])

    @webapi_test_template
    def test_put_finalize_adds_default_reviewers_first_time_only(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 doesn't
        add default reviewers a second time
        """
        @self.spy_for(Repository.get_file_exists,
                      owner=Repository)
        def _get_file_exists(repository, path, revision, *args, **kwargs):
            self.assertEqual(path, 'README')
            self.assertEqual(revision, '94bdd3e')

            return True

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)

            # Create the state needed for the default reviewer.
            group = self.create_review_group(name='group1')

            default_reviewer = DefaultReviewer.objects.create(
                name='default1',
                file_regex='.')
            default_reviewer.groups.add(group)
            default_reviewer.repository.add(review_request.repository)

            # Create the initial diffset. This should prevent a default
            # reviewer from being applied, since we're not publishing the first
            # diff on a review request.
            self.create_diffset(review_request=review_request)

            # Create the state needed for the diff to post.
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)

            commit = self.create_diffcommit(diffset=diffset)

            filediff = FileDiff.objects.get()

            cumulative_diff = SimpleUploadedFile('diff', filediff.diff,
                                                 content_type='text/x-patch')

            validation_info = update_validation_info(
                {},
                commit_id=commit.commit_id,
                parent_id=commit.parent_id,
                filediffs=[filediff])

            # Post the diff.
            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info(
                        validation_info),
                },
                expected_mimetype=diff_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.compare_item(rsp['diff'], diffset)

            draft = review_request.get_draft()
            self.assertEqual(list(draft.target_groups.all()), [])

    @webapi_test_template
    def test_put_finalize_again(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 when the
        diff is already finalized
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            self.create_diffcommit(diffset=diffset)

            diffset.finalize_commit_series(
                cumulative_diff=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                validation_info=None,
                validate=False,
                save=True)

            cumulative_diff = SimpleUploadedFile('diff', b'',
                                                 content_type='text/x-patch')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info({}),
                },
                expected_status=400)

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_ATTRIBUTE.code,
                'msg': INVALID_ATTRIBUTE.msg,
                'type': INVALID_ATTRIBUTE.error_type,
            },
            'reason': 'This diff is already finalized.',
        })

    @webapi_test_template
    def test_put_finalize_missing_fields(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 with missing
        request fields
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            self.create_diffcommit(diffset=diffset)

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                },
                expected_status=400)

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_FORM_DATA.code,
                'msg': INVALID_FORM_DATA.msg,
                'type': INVALID_FORM_DATA.error_type,
            },
            'fields': {
                'cumulative_diff': [
                    'This field is required when finalize_commit_series is '
                    'set.',
                ],
                'validation_info': [
                    'This field is required when finalize_commit_series is '
                    'set.',
                ],
            },
        })

    @webapi_test_template
    def test_put_finalize_review_request_without_history(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 when the
        review request was created without commit history support
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)

            cumulative_diff = SimpleUploadedFile('diff', b'',
                                                 content_type='text/x-patch')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info({}),
                },
                expected_status=400)

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_ATTRIBUTE.code,
                'msg': INVALID_ATTRIBUTE.msg,
                'type': INVALID_ATTRIBUTE.error_type,
            },
            'reason': 'This review request was not created with commit '
                      'history support.',
        })

    @webapi_test_template
    def test_put_finalize_dvcs_feature_disabled(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 when the
        DVCS feature is disabled
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=False):
            review_request = self.create_review_request(
                create_repository=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)

            cumulative_diff = SimpleUploadedFile('diff', b'',
                                                 content_type='text/x-patch')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info({}),
                },
                expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.compare_item(rsp['diff'], diffset)

    @webapi_test_template
    def test_put_finalize_empty_commit_series(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 for an empty
        commit series
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)

            cumulative_diff = SimpleUploadedFile('diff', b'',
                                                 content_type='text/x-patch')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info({}),
                },
                expected_status=400)

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_ATTRIBUTE.code,
                'msg': INVALID_ATTRIBUTE.msg,
                'type': INVALID_ATTRIBUTE.error_type,
            },
            'reason': 'Cannot finalize an empty commit series.',
        })

    @webapi_test_template
    def test_put_finalize_invalid_validation_info_base64(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 when
        validation_info is invalid base64
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            self.create_diffcommit(diffset=diffset)

            cumulative_diff = SimpleUploadedFile('diff', b'',
                                                 content_type='text/x-patch')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': 'foo',
                },
                expected_status=400)

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_FORM_DATA.code,
                'msg': INVALID_FORM_DATA.msg,
                'type': INVALID_FORM_DATA.error_type,
            },
            'fields': {
                'validation_info': [
                    'Could not parse field: Incorrect padding',
                ],
            },
        })

    @webapi_test_template
    def test_put_finalize_invalid_validation_info_json_format(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 when
        validation_info is invalid json
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            self.create_diffcommit(diffset=diffset)

            cumulative_diff = SimpleUploadedFile('diff', b'',
                                                 content_type='text/x-patch')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info('foo'),
                },
                expected_status=400)

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_FORM_DATA.code,
                'msg': INVALID_FORM_DATA.msg,
                'type': INVALID_FORM_DATA.error_type,
            },
            'fields': {
                'validation_info': [
                    'Could not parse field: Invalid format.',
                ],
            },
        })

    @webapi_test_template
    def test_put_finalize_invalid_validation_info_not_json(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 when
        validation_info is JSON in the incorrect format
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            self.create_diffcommit(diffset=diffset)

            cumulative_diff = SimpleUploadedFile('diff', b'',
                                                 content_type='text/x-patch')
            validation_info = base64.b64encode(b'AAAAAAA').decode('utf-8')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': validation_info,
                },
                expected_status=400)

        expected_error = (
            'Could not parse field: Expecting value: line 1 '
            'column 1 (char 0)'
        )

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_FORM_DATA.code,
                'msg': INVALID_FORM_DATA.msg,
                'type': INVALID_FORM_DATA.error_type,
            },
            'fields': {
                'validation_info': [expected_error],
            },
        })

    @webapi_test_template
    def test_put_finalize_validation_info_extra_commits(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 when
        validation_info contains commits that do not exist
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            commit = self.create_diffcommit(diffset=diffset)
            filediff = commit.files.first()

            validation_info = update_validation_info(
                {},
                commit_id=commit.commit_id,
                parent_id=commit.parent_id,
                filediffs=[filediff])

            validation_info = update_validation_info(
                validation_info,
                commit_id='f00',
                parent_id=commit.commit_id,
                filediffs=[])

            cumulative_diff = SimpleUploadedFile('diff', filediff.diff,
                                                 content_type='text/x-patch')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info(
                        validation_info),
                },
                expected_status=400)

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_FORM_DATA.code,
                'msg': INVALID_FORM_DATA.msg,
                'type': INVALID_FORM_DATA.error_type,
            },
            'fields': {
                'validation_info': [
                    'The following commits are specified in validation_info '
                    'but do not exist: f00'
                ],
            },
        })

    @webapi_test_template
    def test_put_finalize_validation_info_missing_commits(self):
        """Testing the PUT <URL> API with finalize_commit_series=1 when
        validation_info does not contain all commits
        """
        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)

            commits = [
                self.create_diffcommit(diffset=diffset, **kwargs)
                for kwargs in (
                    {'commit_id': 'r1', 'parent_id': 'r0'},
                    {'commit_id': 'r2', 'parent_id': 'r1'},
                )
            ]

            filediff = commits[0].files.first()

            validation_info = update_validation_info(
                {},
                commit_id=commits[0].commit_id,
                parent_id=commits[0].parent_id,
                filediffs=[filediff])

            cumulative_diff = SimpleUploadedFile('diff', filediff.diff,
                                                 content_type='text/x-patch')

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info(
                        validation_info),
                },
                expected_status=400)

        self.assertEqual(rsp, {
            'stat': 'fail',
            'err': {
                'code': INVALID_FORM_DATA.code,
                'msg': INVALID_FORM_DATA.msg,
                'type': INVALID_FORM_DATA.error_type,
            },
            'fields': {
                'validation_info': [
                    'The following commits exist but are not present in '
                    'validation_info: r2',
                ],
            }
        })

    @webapi_test_template
    def test_put_finalize_emits_review_request_diffset_uploaded(self) -> None:
        """Testing the PUT <URL> API with finalize_commit_series=1 emits a
        review_request_diffset_uploaded signal
        """
        @self.spy_for(Repository.get_file_exists,
                      owner=Repository)
        def _get_file_exists(repository, path, revision, *args, **kwargs):
            self.assertEqual(path, 'README')
            self.assertEqual(revision, '94bdd3e')

            return True

        def on_diffset_uploaded(diffset, review_request_draft, **kwargs):
            pass

        review_request_diffset_uploaded.connect(on_diffset_uploaded)
        self.spy_on(on_diffset_uploaded)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)
            commit = self.create_diffcommit(diffset=diffset)

            filediff = FileDiff.objects.get()

            cumulative_diff = SimpleUploadedFile('diff', filediff.diff,
                                                 content_type='text/x-patch')

            validation_info = update_validation_info(
                {},
                commit_id=commit.commit_id,
                parent_id=commit.parent_id,
                filediffs=[filediff])

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': True,
                    'cumulative_diff': cumulative_diff,
                    'validation_info': serialize_validation_info(
                        validation_info),
                },
                expected_mimetype=diff_item_mimetype)

            draft = review_request.get_draft()
            diffset = DiffSet.objects.get(pk=rsp['diff']['id'])

            self.assertEqual(rsp['stat'], 'ok')
            self.assertSpyCalledWith(
                on_diffset_uploaded,
                diffset=diffset,
                review_request_draft=draft)

        review_request_diffset_uploaded.disconnect(on_diffset_uploaded)

    @webapi_test_template
    def test_put_not_finalize_review_request_diffset_uploaded(self) -> None:
        """Testing the PUT <URL> API with finalize_commit_series=0 does not
        emit a review_request_diffset_uploaded signal
        """
        def on_diffset_uploaded(diffset, review_request_draft, **kwargs):
            pass

        review_request_diffset_uploaded.connect(on_diffset_uploaded)
        self.spy_on(on_diffset_uploaded)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            review_request = self.create_review_request(
                create_repository=True,
                create_with_history=True,
                submitter=self.user)
            diffset = self.create_diffset(review_request=review_request,
                                          draft=True)

            rsp = self.api_put(
                get_draft_diff_item_url(review_request, diffset.revision),
                {
                    'finalize_commit_series': False,
                    'extra_data': {
                        'foo': 'bar',
                    },
                },
                expected_mimetype=diff_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertSpyNotCalled(on_diffset_uploaded)

        review_request_diffset_uploaded.disconnect(on_diffset_uploaded)



"""Managers for the hostingsvcs app."""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from django.db.models import Manager, Q

from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from django.db.models import QuerySet

    from reviewboard.hostingsvcs.models import HostingServiceAccount
    from reviewboard.site.models import AnyOrAllLocalSites


class HostingServiceAccountManager(Manager['HostingServiceAccount']):
    """A manager for HostingServiceAccount models."""

    def accessible(
        self,
        visible_only: bool = True,
        local_site: Optional[AnyOrAllLocalSites] = None,
    ) -> QuerySet[HostingServiceAccount]:
        """Return hosting service accounts that are accessible.

        These will include all visible accounts that are compatible with the
        specified :term:`Local Site`.

        Version Changed:
            6.0:
            Removed the ``filter_local_site`` argument.

        Version Changed:
            5.0:
            Deprecated ``filter_local_site`` and added support for
            setting ``local_site`` to :py:class:`LocalSite.ALL
            <reviewboard.site.models.LocalSite.ALL>`.

        Args:
            visible_only (bool, optional):
                Whether to only include visible accounts in the results.

            local_site (reviewboard.site.models.AnyOrAllLocalSites, optional):
                A :term:`Local Site` that the accounts must be associated with.
                If not specified, returned accounts won't be bound to a
                Local Site.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

                Version Changed:
                    5.0:
                    Added support for :py:attr:`LocalSite.ALL
                    <reviewboard.site.models.LocalSite.ALL>`.

        Returns:
            django.db.models.query.QuerySet:
            The resulting queryset.
        """
        q = LocalSite.objects.build_q(local_site=local_site)

        if visible_only:
            q &= Q(visible=True)

        return self.filter(q)

    def can_create(
        self,
        user: User,
        local_site: Optional[AnyOrAllLocalSites] = None,
    ) -> bool:
        """Return whether the user can create a hosting service account.

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

            local_site (reviewboard.site.models.AnyOrAllLocalSites, optional):
                A :term:`Local Site` to create hosting service accounts on.

                This may be :py:attr:`LocalSite.ALL
                <reviewboard.site.models.LocalSite.ALL>`.

        Returns:
            bool:
            True if the user can create hosting service accounts.
        """
        return user.has_perm('hostingsvcs.create_hostingserviceaccount',
                             local_site)



from django.urls import include, path, re_path

from reviewboard.datagrids import views


urlpatterns = [
    # All Review Requests
    path('r/',
         views.all_review_requests,
         name='all-review-requests'),

    # Dashboard
    path('dashboard/',
         views.dashboard,
         name='dashboard'),

    # Users
    path('users/', include([
        path('',
             views.users_list,
             name='all-users'),
        re_path(r'^(?P<username>[\w.@+-]+)/$',
                views.submitter,
                name='user'),
        re_path(r'^(?P<username>[\w.@+-]+)/(?P<grid>[a-z-]+)/$',
                views.submitter,
                name='user-grid'),
    ])),

    # Groups
    path('groups/', include([
        path('',
             views.group_list,
             name='all-groups'),
        path('<str:name>/',
             views.group,
             name='group'),
        path('<str:name>/members/',
             views.group_members,
             name='group-members'),
    ])),
]



"""Unit tests for reviewboard.reviews.models.Review."""

from datetime import datetime, timedelta, timezone

from django.contrib.auth.models import AnonymousUser, User
from django.utils import timezone as django_timezone
from djblets.testing.decorators import add_fixtures
from djblets.util.dates import get_tz_aware_utcnow
from kgb import SpyAgency, spy_on

from reviewboard.reviews.errors import RevokeShipItError
from reviewboard.reviews.models import (Comment,
                                        FileAttachmentComment,
                                        Review,
                                        ReviewRequest)
from reviewboard.reviews.signals import (review_ship_it_revoked,
                                         review_ship_it_revoking)
from reviewboard.testing import TestCase


class ReviewTests(SpyAgency, TestCase):
    """Unit tests for reviewboard.reviews.models.Review."""

    fixtures = ['test_users', 'test_scmtools']

    def test_all_participants_with_replies(self):
        """Testing Review.all_participants with replies"""
        user1 = User.objects.create_user(username='aaa',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='bbb',
                                         email='user2@example.com')
        user3 = User.objects.create_user(username='ccc',
                                         email='user3@example.com')
        user4 = User.objects.create_user(username='ddd',
                                         email='user4@example.com')

        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, user=user1)
        self.create_reply(review, user=user2, public=True)
        self.create_reply(review, user=user1, public=True)
        self.create_reply(review, user=user4, public=False)
        self.create_reply(review, user=user3, public=True)
        self.create_reply(review, user=user2, public=True)

        with self.assertNumQueries(2):
            self.assertEqual(review.all_participants, {user1, user2, user3})

    def test_all_participants_with_no_replies(self):
        """Testing Review.all_participants with no replies"""
        user = User.objects.create_user(username='aaa',
                                        email='user1@example.com')

        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, user=user)

        with self.assertNumQueries(1):
            self.assertEqual(review.all_participants, {user})

    def test_all_participants_with_only_owner_reply(self):
        """Testing Review.all_participants with only review owner replied"""
        user = User.objects.create_user(username='aaa',
                                        email='user1@example.com')

        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, user=user)
        self.create_reply(review, user=user, public=True)

        with self.assertNumQueries(1):
            self.assertEqual(review.all_participants, {user})

    def test_is_accessible_by_with_public_and_anonymous_user(self):
        """Testing Review.is_accessible_by with public and anonymous user"""
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request,
                                    user=user,
                                    public=True)

        self.assertTrue(review.is_accessible_by(AnonymousUser()))

    def test_is_accessible_by_with_public_and_private_review_request(self):
        """Testing Review.is_accessible_by with public review and private
        review request
        """
        user = User.objects.get(username='doc')
        other_user = User.objects.get(username='dopey')

        review_request = self.create_review_request(create_repository=True)
        review = self.create_review(review_request,
                                    user=user,
                                    public=True)

        review_request.repository.public = True
        review_request.repository.save(update_fields=('public',))

        self.assertFalse(review.is_accessible_by(other_user))

    def test_is_accessible_by_with_private_and_anonymous_user(self):
        """Testing Review.is_accessible_by with private and anonymous user"""
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request,
                                    user=user)

        self.assertFalse(review.is_accessible_by(AnonymousUser()))

    def test_is_accessible_by_with_private_and_owner(self):
        """Testing Review.is_accessible_by with private and owner"""
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request,
                                    user=user)

        self.assertTrue(review.is_accessible_by(user))

    def test_is_accessible_by_with_private_and_superuser(self):
        """Testing Review.is_accessible_by with private and superuser"""
        user = User.objects.get(username='doc')
        admin = User.objects.get(username='admin')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request,
                                    user=user)

        self.assertTrue(review.is_accessible_by(admin))

    def test_is_mutable_by_with_public_and_owner(self):
        """Testing Review.is_mutable_by with public and owner"""
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request,
                                    user=user,
                                    public=True)

        self.assertFalse(review.is_mutable_by(user))

    def test_is_mutable_by_with_private_and_anonymous_user(self):
        """Testing Review.is_mutable_by with private and anonymous user"""
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request,
                                    user=user)

        self.assertFalse(review.is_mutable_by(AnonymousUser()))

    def test_is_mutable_by_with_private_and_owner(self):
        """Testing Review.is_mutable_by with private and owner"""
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request,
                                    user=user)

        self.assertTrue(review.is_mutable_by(user))

    def test_is_mutable_by_with_private_and_superuser(self):
        """Testing Review.is_mutable_by with private and superuser"""
        user = User.objects.get(username='doc')
        admin = User.objects.get(username='admin')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request,
                                    user=user)

        self.assertTrue(review.is_mutable_by(admin))

    def test_is_new_for_user_with_non_owner(self):
        """Testing Review.is_new_for_user with non-owner"""
        user1 = User.objects.create_user(username='test-user-1',
                                         email='user1@example.com')
        user2 = User.objects.create_user(username='test-user-2',
                                         email='user2@example.com')

        review = Review(
            user=user1,
            timestamp=datetime(2017, 9, 7, 15, 27, 0))
        self.assertTrue(review.is_new_for_user(
            user=user2,
            last_visited=datetime(2017, 9, 7, 10, 0, 0)))
        self.assertFalse(review.is_new_for_user(
            user=user2,
            last_visited=datetime(2017, 9, 7, 16, 0, 0)))
        self.assertFalse(review.is_new_for_user(
            user=user2,
            last_visited=datetime(2017, 9, 7, 15, 27, 0)))

    def test_is_new_for_user_with_owner(self):
        """Testing Review.is_new_for_user with owner"""
        user = User.objects.create_user(username='test-user',
                                        email='user@example.com')

        review = Review(
            user=user,
            timestamp=datetime(2017, 9, 7, 15, 27, 0))
        self.assertFalse(review.is_new_for_user(
            user=user,
            last_visited=datetime(2017, 9, 7, 16, 0, 0)))

    def test_can_user_revoke_ship_it_with_owner(self):
        """Testing Review.can_user_revoke_ship_it with review owner"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        self.assertTrue(review.can_user_revoke_ship_it(review.user))

    def test_can_user_revoke_ship_it_with_non_owner(self):
        """Testing Review.can_user_revoke_ship_it with non-owner"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        user = User.objects.get(username='doc')
        self.assertNotEqual(review.user, user)

        self.assertFalse(review.can_user_revoke_ship_it(user))

    def test_can_user_revoke_ship_it_with_superuser(self):
        """Testing Review.can_user_revoke_ship_it with superuser"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        user = User.objects.get(username='admin')
        self.assertNotEqual(review.user, user)

        self.assertTrue(review.can_user_revoke_ship_it(user))

    @add_fixtures(['test_site'])
    def test_can_user_revoke_ship_it_with_local_site_admin(self):
        """Testing Review.can_user_revoke_ship_it with LocalSite admin"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True,
                                                    with_local_site=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        user = User.objects.create_user(username='new-site-admin',
                                        email='new_site_admin@example.com')
        review_request.local_site.admins.add(user)
        review_request.local_site.users.add(user)

        self.assertTrue(review.can_user_revoke_ship_it(user))

    def test_can_user_revoke_ship_it_with_anonymous(self):
        """Testing Review.can_user_revoke_ship_it with anonymous user"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        self.assertFalse(review.can_user_revoke_ship_it(AnonymousUser()))

    def test_can_user_revoke_ship_it_with_unpublished(self):
        """Testing Review.can_user_revoke_ship_it with unpublished review"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True)

        self.assertFalse(review.can_user_revoke_ship_it(review.user))

    def test_can_user_revoke_ship_it_with_no_ship_it(self):
        """Testing Review.can_user_revoke_ship_it with no Ship It"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request)

        self.assertFalse(review.can_user_revoke_ship_it(review.user))

    def test_revoke_ship_it(self):
        """Testing Review.revoke_ship_it"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        self.spy_on(review_ship_it_revoking.send)
        self.spy_on(review_ship_it_revoked.send)

        self.assertEqual(review_request.shipit_count, 1)

        review.revoke_ship_it(review.user)

        # Make sure the signals fired.
        self.assertTrue(review_ship_it_revoking.send.called_with(
            sender=Review, user=review.user, review=review))
        self.assertTrue(review_ship_it_revoked.send.called_with(
            sender=Review, user=review.user, review=review))

        # Check the state of the fields.
        self.assertEqual(review.body_top, Review.REVOKED_SHIP_IT_TEXT)
        self.assertFalse(review.ship_it)
        self.assertTrue(review.extra_data.get('revoked_ship_it'))
        self.assertEqual(review_request.shipit_count, 0)

        # Make sure they persisted to the database.
        review = Review.objects.get(pk=review.pk)
        self.assertEqual(review.body_top, Review.REVOKED_SHIP_IT_TEXT)
        self.assertFalse(review.ship_it)
        self.assertTrue(review.extra_data.get('revoked_ship_it'))

    def test_revoke_ship_it_with_no_ship_it(self):
        """Testing Review.revoke_ship_it with no Ship It"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    publish=True)

        expected_error = 'This review is not marked Ship It!'

        with self.assertRaisesMessage(RevokeShipItError, expected_error):
            review.revoke_ship_it(review.user)

        self.assertEqual(review.body_top, Review.SHIP_IT_TEXT)
        self.assertFalse(review.ship_it)

    def test_revoke_ship_it_with_custom_body_top(self):
        """Testing Review.revoke_ship_it with custom existing body_top"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top='This is a test',
                                    ship_it=True,
                                    publish=True)

        review.revoke_ship_it(review.user)

        self.assertEqual(review.body_top, 'This is a test')
        self.assertFalse(review.ship_it)
        self.assertTrue(review.extra_data.get('revoked_ship_it'))

    def test_revoke_ship_it_with_revoking_signal_exception(self):
        """Testing Review.revoke_ship_it with exception in
        review_ship_it_revoking handler
        """
        def on_revoking(**kwargs):
            raise Exception('oh no')

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        try:
            review_ship_it_revoking.connect(on_revoking)

            expected_error = 'Error revoking the Ship It: oh no'

            with self.assertRaisesMessage(RevokeShipItError, expected_error):
                review.revoke_ship_it(review.user)
        finally:
            review_ship_it_revoking.disconnect(on_revoking)

        self.assertEqual(review.body_top, Review.SHIP_IT_TEXT)
        self.assertTrue(review.ship_it)
        self.assertNotIn('revoked_ship_it', review.extra_data)

    def test_revoke_ship_it_with_revoked_signal_exception(self):
        """Testing Review.revoke_ship_it with exception in
        review_ship_it_revoked handler
        """
        def on_revoked(**kwargs):
            raise Exception('oh no')

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True,
                                    publish=True)

        try:
            review_ship_it_revoked.connect(on_revoked)
            review.revoke_ship_it(review.user)
        finally:
            review_ship_it_revoked.disconnect(on_revoked)

        self.assertEqual(review.body_top, Review.REVOKED_SHIP_IT_TEXT)
        self.assertFalse(review.ship_it)
        self.assertTrue(review.extra_data.get('revoked_ship_it'))

    def test_revoke_ship_it_timestamp(self):
        """Testing Review.revoke_ship_it does not modify the review timestamp
        """
        # ReviewRequest.last_update is a
        # django.db.fields.ModificationTimestampField, which retrieves its
        # value from datetime.utcnow().replace(tzinfo=utc).
        #
        # django.utils.timezone.now has the same implementation.
        #
        # Unfortunately, we cannot spy on datetime.utcnow since it is a
        # builtin. So we replace get_tz_aware_utcnow with timezone.now and we
        # will replace that with a constant function in the spy_on calls below.
        self.spy_on(get_tz_aware_utcnow,
                    call_fake=lambda: django_timezone.now())

        creation_timestamp = datetime.fromtimestamp(0, timezone.utc)
        review_timestamp = creation_timestamp + timedelta(hours=1)
        revoke_timestamp = review_timestamp + timedelta(hours=1)

        with spy_on(django_timezone.now, call_fake=lambda: creation_timestamp):
            review_request = self.create_review_request(publish=True)

        with spy_on(django_timezone.now, call_fake=lambda: review_timestamp):
            review = self.create_review(review_request,
                                        body_top=Review.SHIP_IT_TEXT,
                                        ship_it=True,
                                        publish=True)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)

        self.assertEqual(review_request.time_added, creation_timestamp)
        self.assertEqual(review_request.last_updated, review_timestamp)
        self.assertEqual(review.timestamp, review_timestamp)

        with spy_on(django_timezone.now, call_fake=lambda: revoke_timestamp):
            review.revoke_ship_it(review.user)

        review = Review.objects.get(pk=review.pk)
        review_request = ReviewRequest.objects.get(pk=review_request.pk)

        self.assertEqual(review_request.time_added, creation_timestamp)
        self.assertEqual(review_request.last_updated, review_timestamp)
        self.assertEqual(review.timestamp, review_timestamp)

    def test_draft_diff_comment_invalidation(self):
        """Testing invalidation of comments on draft diffs when the diff is
        replaced.
        """
        review_request = self.create_review_request(create_repository=True)
        diffset = self.create_diffset(review_request=review_request,
                                      draft=True)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request)
        comment = self.create_diff_comment(review, filediff)

        # DiffResource.create() will delete an existing draft DiffSet before
        # replacing it, so we can simulate that here by just deleting the one
        # we have.
        diffset.delete()

        self.assertFalse(Comment.objects.filter(pk=comment.pk).exists())

    def test_draft_file_attachment_comment_invalidation(self):
        """Testing invalidation of comments on draft file attachments when the
        attachment is replaced.
        """
        review_request = self.create_review_request(create_repository=True)
        attachment = self.create_file_attachment(review_request,
                                                 draft=True)

        review = self.create_review(review_request)
        comment = self.create_file_attachment_comment(review, attachment)

        # UploadAttachmentForm.create() will delete an existing draft
        # attachment before replacing it, so we can simulate that here by just
        # deleting the one we have.
        attachment.delete()

        self.assertFalse(
            FileAttachmentComment.objects.filter(pk=comment.pk).exists())

    def test_can_publish(self):
        """Testing Review.can_publish"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True)

        can_publish, err = review.can_publish()

        self.assertTrue(can_publish)
        self.assertIsNone(err)

    def test_can_publish_draft_review_request(self):
        """Testing Review.can_publish with a draft review request"""
        review_request = self.create_review_request(create_repository=True)
        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True)

        can_publish, err = review.can_publish()

        self.assertFalse(can_publish)
        self.assertEqual(err,
                         'This review cannot be published until the review '
                         'request is published.')

    def test_can_publish_with_draft_diff_comments(self):
        """Testing Review.can_publish with comments on a draft diff"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(diffset)

        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True)
        self.create_diff_comment(review, filediff)

        can_publish, err = review.can_publish()

        self.assertFalse(can_publish)
        self.assertEqual(err,
                         'This review cannot be published, because it '
                         'includes a comment on a diff which has not yet '
                         'been published.')

    def test_can_publish_with_draft_interdiff_comments(self):
        """Testing Review.can_publish with comments on a draft interdiff"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)

        interdiffset = self.create_diffset(review_request, draft=True)
        interfilediff = self.create_filediff(interdiffset)

        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True)
        self.create_diff_comment(review, filediff, interfilediff)

        can_publish, err = review.can_publish()

        self.assertFalse(can_publish)
        self.assertEqual(err,
                         'This review cannot be published, because it '
                         'includes a comment on a diff which has not yet '
                         'been published.')

    def test_can_publish_with_draft_file_comments(self):
        """Testing Review.can_publish with comments on a draft file attachment
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        attachment = self.create_file_attachment(review_request, draft=True)

        review = self.create_review(review_request,
                                    body_top=Review.SHIP_IT_TEXT,
                                    ship_it=True)
        self.create_file_attachment_comment(review, attachment)

        can_publish, err = review.can_publish()

        self.assertFalse(can_publish)
        self.assertEqual(err,
                         'This review cannot be published, because it '
                         'includes a comment on a file attachment which has '
                         'not yet been published.')



"""Unit tests for reviewboard.datagrids.columns.PendingCountColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import PendingCountColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class PendingCountColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.PendingCountColumn.

    Version Added:
        8.0
    """

    column = PendingCountColumn()

    def test_render_data_with_zero_pending(self) -> None:
        """Testing PendingCountColumn.render_data with zero pending"""
        group = self.create_review_group()

        # This is normally set by the column's augment_queryset_for_data().
        group.column_pending_review_request_count = 0

        value = self.column.render_data(self.stateful_column, group)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '0')

    def test_render_data_with_pending(self) -> None:
        """Testing PendingCountColumn.render_data with pending"""
        group = self.create_review_group()

        # This is normally set by the column's augment_queryset_for_data().
        group.column_pending_review_request_count = 4

        value = self.column.render_data(self.stateful_column, group)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '4')

    def test_to_json_with_zero_pending(self) -> None:
        """Testing PendingCountColumn.to_json with zero pending"""
        group = self.create_review_group()

        # This is normally set by the column's augment_queryset_for_data().
        group.column_pending_review_request_count = 0

        self.assertEqual(
            self.column.to_json(self.stateful_column, group),
            0)

    def test_to_json_with_pending(self) -> None:
        """Testing PendingCountColumn.to_json with pending"""
        group = self.create_review_group()

        # This is normally set by the column's augment_queryset_for_data().
        group.column_pending_review_request_count = 4

        self.assertEqual(
            self.column.to_json(self.stateful_column, group),
            4)
